from .Hand import *

class InspireHand(Hand):
    def __init__(self, ip, timeout=0.1):
        super().__init__(ip, timeout = timeout)
        self.kSend_Frame_Head1 = 0xEB
        self.kSend_Frame_Head2 = 0x90
        self.ID = 1
        self.kRcv_Frame_Head1 = 0x90
        self.kRcv_Frame_Head2 = 0xEB
        self.kCmd_Handg3_Read = 0x11  
        self.kCmd_Handg3_Write = 0x12


    def Send_Messages(self, tx_messages):
        """
        send message using udp socket for inspire hand
        
        Args : tx_messages, list of bytes

        Returns:
            FunctionResult in dexhand.py        
        """ 
        try:
            ret = self.s.SendData(tx_messages, self.ctrl_port)
            rec_data = self.s.ReceiveData(1024)[1]
            if ret != FunctionResult.SUCCESS:
                raise Exception("SendData failed")
            if rec_data is None or tx_messages[4:6] != rec_data[4:6]:
                raise Exception("ReceiveData failed")
            return FunctionResult.SUCCESS
        except Exception as e:
            logger.print_trace_error("send_message failed: {}".format(e))
            return FunctionResult.FAIL


    def Get_Messages(self, tx_messages, type = "6short"):
        """
        get message using udp socket from inspire hand
        
        Args : tx_messages, list of bytes

        Returns:
            FunctionResult in dexhand.py        
        """ 
        try:
            ret = self.s.SendData(tx_messages, self.ctrl_port)
            rec_data = self.s.ReceiveData(1024)[1]
            if ret != FunctionResult.SUCCESS:
                raise Exception("SendData failed")
            if rec_data is None or tx_messages[4:6] != rec_data[4:6]:
                raise Exception("ReceiveData failed")
            integers = []
            if type == "6short":
                for i in range(7, len(rec_data)-1, 2):  
                    value = struct.unpack_from('<h', rec_data, i)[0]
                    integers.append(value)
            elif type == "6byte":
                for i in range(7, len(rec_data)-1, 1):  
                    value = struct.unpack_from('<B', rec_data, i)[0]
                    integers.append(value)
            else:
                raise ValueError("Unsupported type: {}".format(type))
            return integers    
        except Exception as e:
            logger.print_trace_error("get_message failed: {}".format(e))
            return None


    def set_ip(self, ip):
        """
        set ip of current or hand
        
        Args : New ip

        Returns:
            FunctionResult in dexhand.py        
        """ 
        data = {
            "method": "SET", "reqTarget" : "/config", "property":"", 
            "static_IP": ip
        }
        json_str = json.dumps(data)
        self.s.SendData(str.encode(json_str), self.comm_port)
        try:
            rec_data = self.s.ReceiveData(1024)[1]
            json_ = json.loads(rec_data.decode("utf-8"))
            if json_["status"] == "OK":
                return FunctionResult.SUCCESS
        except:
            return FunctionResult.FAIL

    def get_ip(self):
        """
        Broadcasting on port comm_port to get ip of inspirehand
        Returns:
            string ip address
        """
        message = b"get ip"
        s = HandSocket("192.168.137.255", timeout=0.1)
        s.s.bind(("", self.comm_port))
        s.SendData(message, self.comm_port)
        data, addr = s.ReceiveData(1024)[1], s.ReceiveData(1024)[2]   
        print(f"Received message from {addr}: {data}")
        return addr


    def clean_error(self):
        """
        clean errors of current or  Electric Cylinder.

        Returns:
            FunctionResult in dexhand.py        
        """
        tx_messages = struct.pack('<BBBBBBBB', 
                          self.kSend_Frame_Head1, 
                          self.kSend_Frame_Head2, 
                          int(self.ID), 
                          0x04, 
                          self.kCmd_Handg3_Write, 
                          0xEC, 
                          0x03,
                          0x01)
        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)  
        return self.Send_Messages(tx_messages)
        

    def save_config_flash(self):
        """
        save parameters to flash.

        Returns:
            FunctionResult in dexhand.py        
        """ 
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Write, 
                    0xED, 
                    0x03,
                    0x01)
        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)  
        return self.Send_Messages(tx_messages)


    def reset(self):
        """
        reset to factory defaults.

        Returns:
            FunctionResult in dexhand.py        
        """ 
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Write, 
                    0xEE, 
                    0x03,
                    0x01)
        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)  
        return self.Send_Messages(tx_messages)


    def calibration(self):
        """
        Force sensor calibration, continue for about 10 seconds.

        Returns:
            FunctionResult in dexhand.py        
        """  
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Write, 
                    0xF1, 
                    0x03,
                    0x01)
        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)  
        return self.Send_Messages(tx_messages)

    

    def set_current_limit(self, current_limit):
        """
        Set the default current limit information when starting.
        
        Args:
            List : [current little finger, current ring finger, ..]
            Limit : 0-1500 mA
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0xFC, 
                    0x03, 
                    int(current_limit[0]),
                    int(current_limit[1]),
                    int(current_limit[2]),
                    int(current_limit[3]),
                    int(current_limit[4]), 
                    int(current_limit[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)


    # [0 -1000]
    def set_default_speed(self, default_speed):
        """
        Set the default speed limit information when starting.
        
        Args:
            List : [default speed little finger, default speed ring finger, ..]
            Limit : 0-1000
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0x08, 
                    0x04, 
                    int(default_speed[0]),
                    int(default_speed[1]),
                    int(default_speed[2]),
                    int(default_speed[3]),
                    int(default_speed[4]), 
                    int(default_speed[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)

 
    def set_default_force(self, default_force):
        """
        Set the default force limit information when starting.
        
        Args:
            List : [default force little finger, default force ring finger, ..]
            Limit : 0-1000
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0x14, 
                    0x04, 
                    int(default_force[0]),
                    int(default_force[1]),
                    int(default_force[2]),
                    int(default_force[3]),
                    int(default_force[4]), 
                    int(default_force[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)


    def set_pos(self, pos):
        """
        Set the position while moving fingers.
        
        Args:
            List : [position little finger, position ring finger, ..]
            Limit : 0-2000 / -1 : no move
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0xC2, 
                    0x05, 
                    int(pos[0]),
                    int(pos[1]),
                    int(pos[2]),
                    int(pos[3]),
                    int(pos[4]), 
                    int(pos[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)

 
    def set_angle(self, angle):
        """
        Set the angle while moving fingers.
        
        Args:
            List : [angle little finger, angle ring finger, ..]
            Limit : 0-1000 / -1 : no move
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0xCE, 
                    0x05, 
                    int(angle[0]),
                    int(angle[1]),
                    int(angle[2]),
                    int(angle[3]),
                    int(angle[4]), 
                    int(angle[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)


    def set_force_limit(self, force):
        """
        Set the force limit information while moving fingers.
        
        Args:
            List : [force little finger, force ring finger, ..]
            Limit : 0-1000
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0xDA, 
                    0x05, 
                    int(force[0]),
                    int(force[1]),
                    int(force[2]),
                    int(force[3]),
                    int(force[4]), 
                    int(force[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)


    # [0-1000]
    def set_speed(self, speed):
        """
        Set the speed status information while moving fingers.
        
        Args:
            List : [speed little finger, speed ring finger, ..]
            Limit : 0-1000
        
        Returns:
            FunctionResult in dexhand.py
            
        """  
        tx_messages = struct.pack('<BBBBBBBhhhhhh', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x0F, 
                    self.kCmd_Handg3_Write, 
                    0xF2, 
                    0x05, 
                    int(speed[0]),
                    int(speed[1]),
                    int(speed[2]),
                    int(speed[3]),
                    int(speed[4]), 
                    int(speed[5]))

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Send_Messages(tx_messages)


    def get_pos(self):
        """
        Get the position status information.

        Returns:
            List : [position little finger, position ring finger, ..]
            Limit : 0-1000
        """  
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0xFE, 
                    0x05,
                    0x0C)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages)


    def get_angle(self):
        """
        Get the angle status information.

        Returns:
            List : [angle little finger, angle ring finger, ..]
            Limit : 0-1000
        """        
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0x0A, 
                    0x06,
                    0x0C)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages)


    def get_force(self):
        """
        Get the force status information.

        Returns:
            List : [force little finger, force ring finger, ..]
            Limit : 0-1500
        """        
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0x2E, 
                    0x06,
                    0x0C)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages)


    def get_current(self):
        """
        Get the current status information.

        Returns:
            List : [force little finger, force ring finger, ..]
            Limit : 0-1500
        """
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0x3A, 
                    0x06,
                    0x0C)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages)


    def get_errorcode(self):
        """
        get error code of 6 fingers
        0 means no error 

        Returns:
            List : [error code little finger, error code ring finger, ..]

        """        
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0x46, 
                    0x06,
                    0x06)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages, type="6byte")


    def get_status(self):
        """
        get status 0-7 of 6 fingers
        0: Releasing 1: Grasping 2: Stop 3: Force Stop 4: Stop due to Current Protection 
        5: Stop due to Electric Cylinder Stall 6: Stop due to Electric Cylinder Malfunction

        Returns:
            List : [status little finger, status ring finger, ..]

        """
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0x4C, 
                    0x06,
                    0x06)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages, type="6byte")


    def get_temperature(self):
        """
        get temperature of 6 fingers

        Args:
            None

        Returns:
            List : [temperature little finger, temperature ring finger, ..]
            Limit : 0-100

        """        
        tx_messages = struct.pack('<BBBBBBBB', 
                    self.kSend_Frame_Head1, 
                    self.kSend_Frame_Head2, 
                    int(self.ID), 
                    0x04, 
                    self.kCmd_Handg3_Read, 
                    0x52, 
                    0x06,
                    0x06)

        checksum = sum(tx_messages[2:]) & 0xFF
        tx_messages += struct.pack('>B', checksum)
        return self.Get_Messages(tx_messages, type="6byte")

 


import setuptools 
 
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
 
setuptools.setup(
    name="fourier-dhx", 
    version="0.0.2",    
    author="Fei liu, Jinglue Hang",    
    author_email="jinglue.hang@fftai.com",    
    description="A smart sdk for fourier dexterous hand example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',    #对python的最低版本要求
)


import sys
import os
import time
import socket
import platform
import logging
import struct, json

class Logger:
    STATE_OFF = 0x00
    STATE_ON = 0x01

    LEVEL_NONE = 0x00
    LEVEL_TRANCE = 0x01
    LEVEL_DEBUG = 0x02
    LEVEL_WARNING = 0x03
    LEVEL_ERROR = 0x04

    def __init__(self, enable_log_file=False):
        self.state = 0x01
        self.level = 0x00

        if enable_log_file:
            self.__init_log()

    def __init_log(self):

        LOG_FORMAT = "%(asctime)s.%(msecs)03d - %(levelname)s - %(message)s"
        DATE_FORMAT = "%Y/%m/%d %H:%M:%S"
        # INFO DEBUG WARNING ERROR CRITICAL
        LOG_LEVEL = logging.INFO

        project_path = os.path.dirname(os.path.abspath(__file__))  # 
        exe_file_path = project_path
        print(exe_file_path)

        sysstr = platform.system()

        if (sysstr == "Windows"):
            self.is_windows_flag = True
        else:
            self.is_windows_flag = False

        if self.is_windows_flag == True:
            path = exe_file_path + "\\log\\"
            name = str("{:02d}\{:02d}\\".format(time.localtime().tm_year, time.localtime().tm_mon))
        else:
            path = './log/'
            name = str("{:02d}/{:02d}/".format(time.localtime().tm_year, time.localtime().tm_mon))

        self.log_dir_path = path + name

        isExists = os.path.exists(self.log_dir_path)
        if not isExists:
            os.makedirs(self.log_dir_path)
            print('Log Created', self.log_dir_path)
        else:
            print('Log Already Existed', self.log_dir_path)
        self.log_name = str(
            "{:02d}_{:02d}_{:02d}_{:02d}.{:02d}.{:02d}".format(time.localtime().tm_year, time.localtime().tm_mon,
                                                               time.localtime().tm_mday, time.localtime().tm_hour,
                                                               time.localtime().tm_min, time.localtime().tm_sec))
        self.log_dir = self.log_dir_path + self.log_name + ".log"

        logging.basicConfig(filename=self.log_dir, level=LOG_LEVEL, format=LOG_FORMAT, datefmt=DATE_FORMAT)
        logging.log(logging.INFO, "\n\n########## The robot is starting !!! ##########\n\n\n")

    def print_log_file_debug(self, str_temp):
        logging.log(logging.DEBUG, str_temp)

    def print_log_file_info(self, str_temp):
        logging.log(logging.INFO, str_temp)

    def print_log_file_warning(self, str_temp):
        logging.log(logging.WARNING, str_temp)

    def print_log_file_error(self, str_temp):
        logging.log(logging.ERROR, str_temp)

    def print_log_file_critical(self, str_temp):
        logging.log(logging.CRITICAL, str_temp)

    # unit='us' 'ms' 's'
    def __time(self, unit='us'):
        if unit == 'ms' or unit == 'us':
            t = time.time()
            tmid = int(t * 1000000 % 1000000)
            tms = int(tmid / 1000 % 1000)
            t_get = time.localtime(t)
            if unit == 'us':
                tus = int(tmid % 1000)
                t_usr = [t_get.tm_year, t_get.tm_mon, t_get.tm_mday, t_get.tm_hour, t_get.tm_min, t_get.tm_sec, tms,
                         tus]
                t_usr_time_str = str(t_usr[0]).rjust(4) + '.' + str(t_usr[1]).rjust(2, '0') + '.' + str(t_usr[2]).rjust(
                    2, '0') + ' ' + str(t_usr[3]).rjust(2, '0') + ':' + str(t_usr[4]).rjust(2, '0') + ':' + str(
                    t_usr[5]).rjust(2, '0') + '.' + str(t_usr[6]).rjust(3, '0') + '.' + str(t_usr[7]).rjust(3, '0')
            else:
                t_usr = [t_get.tm_year, t_get.tm_mon, t_get.tm_mday, t_get.tm_hour, t_get.tm_min, t_get.tm_sec, tms]
                t_usr_time_str = str(t_usr[0]).rjust(4) + '.' + str(t_usr[1]).rjust(2, '0') + '.' + str(t_usr[2]).rjust(
                    2, '0') + ' ' + str(t_usr[3]).rjust(2, '0') + ':' + str(t_usr[4]).rjust(2, '0') + ':' + str(
                    t_usr[5]).rjust(2, '0') + '.' + str(t_usr[6]).rjust(3, '0')
            return '[' + t_usr_time_str + ']'
        else:
            now = '[' + time.strftime("%Y-%m-%d %H:%M:%S") + ']'
        return now

    def print(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            print(*objects, sep=sep, end=end, file=file, flush=flush)

    def print_line(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            print(*objects, sep=sep, end=end, file=file, flush=flush)


    def print_trace(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            if self.level <= Logger.LEVEL_TRANCE:
                now = self.__time()
                print(now, "\033[0m Info:   \033[0m", end=' ')
                print(*objects, sep=sep, end=end, file=file, flush=flush)

    def print_trace_debug(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            if self.level <= Logger.LEVEL_DEBUG:
                now = self.__time()
                print(now, "\033[0;31m Debug:  \033[0m", end=' ')
                print(*objects, sep=sep, end=end, file=file, flush=flush)

    def print_trace_warning(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            if self.level <= Logger.LEVEL_WARNING:
                now = self.__time()
                print(now, "\033[0;34;43m Warning:\033[0m", end=' ')
                print(*objects, sep=sep, end=end, file=file, flush=flush)

    def print_trace_error(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            if self.level <= Logger.LEVEL_ERROR:
                now = self.__time()
                print(now, "\033[0;32;41m Error:  \033[0m", end=' ')
                print(*objects, sep=sep, end=end, file=file, flush=flush)

    def print_file(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            print(*objects, sep=sep, end=end, file=file, flush=flush)

    def print_file_trace(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        if self.state == Logger.STATE_ON:
            print(*objects, sep=sep, end=end, file=file, flush=flush)

logger = Logger()

class FunctionResult:
    SUCCESS = 0
    FAIL = -1
    RUNNING = 1
    PREPARE = 2
    EXECUTE = 3
    NOT_EXECUTE = 4
    TIMEOUT = 5

class SocketResult:
    SOCKET_SEND_FAILED = -10001,
    SOCKET_SEND_SIZE_WRONG = -10002,
    SOCKET_RECEIVE_FAILED = -10003,
    SOCKET_RECEIVE_SIZE_WRONG = -10004,


class HandSocket:
    def __init__(self, ip, timeout = 0.1):
        self.ip = ip
        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.s.settimeout(timeout)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        
    def SendData(self, send_data, port):
        ret = self.s.sendto(send_data, (self.ip, port))
        if ret < 0:
            return SocketResult.SOCKET_SEND_FAILED
        if ret != len(send_data):
            return SocketResult.SOCKET_SEND_SIZE_WRONG
        return FunctionResult.SUCCESS
    
    def ReceiveData(self, size):
        try:
            data, address = self.s.recvfrom(size)
            return FunctionResult.SUCCESS, data, address
        except socket.timeout:
            return FunctionResult.TIMEOUT, None, None
        except:
            return FunctionResult.FAIL, None, None


class Hand:
    def __init__(self, ip, ctrl_port = 2333, comm_port = 2334, timeout = 0.1):
        self.ctrl_port = ctrl_port
        self.comm_port = comm_port
        self.s = HandSocket(ip, timeout)
    
    def Send_Messages(self, tx_messages):
        pass

    def Get_Messages(self, tx_messages):
        pass

    def clean_error(self):
        pass

    def save_config_flash(self):
        pass

    def reset(self):
        pass

    def calibration(self):
        pass

    def set_ip(self, ip):
        pass

    
    def get_ip(self):
        pass


    def set_current_limit(self, current_limit):
        pass
    
    def set_force_limit(self, force_limit):
        pass

    def set_speed(self, speed):
        pass
    
    def set_pos(self, pos):
        pass

    def set_angle(self, angle):
        pass

    def get_pos(self):
        pass

    def get_angle(self):
        pass

    def get_force(self):
        pass

    def get_current(self):
        pass

    def get_errorcode(self):
        pass

    def get_status(self):
        pass

    def get_temperature(self):
        pass






    
        
    


from .Hand import *

class DexHand(Hand):
    def __init__(self, ip, timeout=0.1):
        super().__init__(ip=ip, timeout=timeout)

    def Send_Messages(self, tx_messages, port):
        """
        send message using udp socket for dexhand
        
        Args : tx_messages, list of bytes

        Returns:
            FunctionResult in dexhand.py        
        """ 
        try:
            ret = self.s.SendData(tx_messages, port)
            if ret != FunctionResult.SUCCESS:
                raise Exception("SendData failed")
            return FunctionResult.SUCCESS
        except Exception as e:
            logger.print_trace_error("send_message failed: {}".format(e))
            return FunctionResult.FAIL
        

    def Get_Messages(self, tx_messages):
        """
        get message using udp socket from dexhand
        
        Args : tx_messages, list of bytes

        Returns:
            FunctionResult in dexhand.py        
        """ 
        try:
            ret = self.s.SendData(tx_messages, self.comm_port)
            rec_data = self.s.ReceiveData(1024)[1]
            if ret != FunctionResult.SUCCESS:
                raise Exception("SendData failed")
            return rec_data    
        except Exception as e:
            logger.print_trace_error("get_message failed: {}".format(e))
            return None 

    def string2int(self, val):
        ret_list = []
        str_data = str(val)
        str_data = str_data[2:]
        while True:
            index = str_data.find(' ')
            if index != -1:
                ele = int(str_data[:index])
                ret_list.append(int(ele))
                str_data = str_data[index+1:]
            else:
                return ret_list

    def string2float(self, val):
        ret_list = []
        str_data = str(val)
        str_data = str_data[2:]
        while True:
            index = str_data.find(' ')
            if index != -1:
                ele = float(str_data[:index])
                ret_list.append(float(ele))
                str_data = str_data[index+1:]
            else:
                return ret_list

    #### get meassage from dexhand ####
    def get_cnt(self):
        """
        Get encoder count params 

        Returns:
            6 fingers count params : [index, middle, ring, little, thumb-1, thumb-2]
        """
        tx_messages = struct.pack('>B', 0x01)
        data = self.Get_Messages(tx_messages)
        return self.string2int(data)


    def get_angle(self):
        """
        GetHandPostion, get joint motor move position params, not joint angle

        Returns:
            6 fingers motor position params: [index, middle, ring, little, thumb-1, thumb-2], defaut position is -0.5
        
        """
        tx_messages = struct.pack('>B', 0x02)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)
        
    
    def get_current(self):
        """
        GetHandCurrent, get hand motor current params
    
        Returns:
            6 fingers motor current params: [index, middle, ring, little, thumb-1, thumb-2]
    
        """
        tx_messages = struct.pack('>B', 0x04)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)


    def get_velocity(self):
        """
        GetHandVelocity, get hand motor velocity params

        Returns:
            6 fingers motor velocity params: [index, middle, ring, little, thumb-1, thumb-2]
        
        """
        tx_messages = struct.pack('>B', 0x03)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)


    def get_errorcode(self):
        """ 
        Get Errorcode of 6 fingers in dexhand
        0 : no error 
        1 : error in position control
        2 : error in velocity control
        3 : error in current control
        Returns:
            list of errorcode : [errorcode index, errorcode middle, ...]
        """ 
        tx_messages = struct.pack('>B', 0x06)
        data = self.Get_Messages(tx_messages)
        return self.string2int(data)


    def get_status(self):
        """
        GetHandStatus, get hand motor status

        Returns:
            6 fingers status params : [index, middle, ring, little, thumb-1, thumb-2], 
      
        """
        tx_messages = struct.pack('>B', 0x05)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)


    def get_pos_limited(self):
        """
        GetHandLimited (Position mode)
        get hand position limited bound, low is 0.0, high is 12.0
        
        Returns:
            6 fingers position high bound params : [index_low, index_high, middle_low, middle_high, ring_low, ring_high, little_low, little_high, thumb-1_low, thumb-1_high, thumb-2_low, thumb-2_high]   
        """
        tx_messages = struct.pack('>B', 0x0b)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)


    def get_velocity_limited(self):
        """
        GetHandLimited (Position mode),get hand position limited bound, low is 0.0, high is 12.0
        Returns:
            FunctionResult in dexhand.py        
        """
        tx_messages = struct.pack('>B', 0x0c)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)
    

    def get_current_limited(self):
        """
        GetHandLimited (Velocity mode), pget hand velocity limited bound, high bound is 10000.0

        Returns:
            6 fingers velocity params : [index, middle, ring, little, thumb-1, thumb-2], 
       
        """
        tx_messages = struct.pack('>B', 0x0d)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)
    

    def get_ip(self):
        """
        Broadcasting on port comm_port to get ip of dexhand
        Returns:
            string ip address
        """
        tx_messages = struct.pack('>B', 0x0f)
        s = HandSocket("192.168.137.255", timeout=0.1)
        s.s.bind(("", self.comm_port))
        s.SendData(tx_messages, self.comm_port)
        _, addr = s.ReceiveData(1024)[1], s.ReceiveData(1024)[2]   
        return addr[0]


    def get_firmware_version(self):
        """
        GetHandFirmwareVersion
        Returns:
            hand version string        
        """
        tx_messages = struct.pack('>B', 0x0e)
        return self.Get_Messages(tx_messages)


    def get_pos_pid(self):
        """ 
        get position mode pid default params : [p, i, d] 
        Returns:
            list of pid : [pid index, pid middle, ...]
        """ 
        tx_messages = struct.pack('>B', 0x08)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)


    def get_velocity_pid(self):
        """ 
        get velocity mode pid default params : [p, i, d] 
        Returns:
            list of pid : [pid index, pid middle, ...]
        """ 
        tx_messages = struct.pack('>B', 0x09)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)


    def get_current_pid(self):
        """ 
        get current mode pid default params : [p, i, d] 
        Returns:
            list of pid : [pid index, pid middle, ...]
        """ 
        tx_messages = struct.pack('>B', 0x0a)
        data = self.Get_Messages(tx_messages)
        return self.string2float(data)



    ### set parameters for dexhand ####

    def set_hand(self, name):
        """ 
        set new dexhand name string < 16byte 
        Returns:
            FunctionResult in dexhand.py 
        """ 
        tx_messages = struct.pack('>B16s', 0x21, name.encode())
        return self.Send_Messages(tx_messages, port=self.comm_port)


    def set_type(self, type = "right"):
        """ 
        set hand type 
        
        Args:
            string left / right 
        
        Returns:
            FunctionResult in dexhand.py 
        """ 
        if type == "right":
            tx_messages = struct.pack('>BB', 0x22, 0x01)
        elif type == "left":
            tx_messages = struct.pack('>BB', 0x22, 0x00)
        return self.Send_Messages(tx_messages, port=self.comm_port)


    def set_ip(self, ip):
        """ 
        set dexhand new ip adress : 192.xxx.xxx.xxx
        Returns:
            FunctionResult in dexhand.py 
        """ 
        try:
            parts = ip.split(".")
            s = HandSocket("192.168.137.255", timeout=0.1)
            s.s.bind(("", self.comm_port))
            tx_messages = struct.pack('>BBBBB', 0x23, *map(int, parts))    
            s.SendData(tx_messages, self.comm_port)
            return FunctionResult.SUCCESS
        except Exception as e:
            print(f"An error occurred: {e}")
            return FunctionResult.FAILURE


    def set_mac(self, mac):
        """ 
        set dexhand new mac (6 byte)
        Returns:
            FunctionResult in dexhand.py 
        """ 
        try:
            parts = mac.split(".")
            s = HandSocket("192.168.137.255", timeout=0.1)
            s.s.bind(("", self.comm_port))
            tx_messages = struct.pack('>BBBBB', 0x24, *map(int, parts))    
            s.SendData(tx_messages, self.comm_port)
            return FunctionResult.SUCCESS
        except Exception as e:
            print(f"An error occurred: {e}")
            return FunctionResult.FAILURE



    def set_SN(self, sn):
        """ 
        set dexhand new SN (6 byte)
        Returns:
            FunctionResult in dexhand.py 
        """ 
        try:
            parts = sn.split(".")
            s = HandSocket("192.168.137.255", timeout=0.1)
            s.s.bind(("", self.comm_port))
            tx_messages = struct.pack('>BBBBB', 0x25, *map(int, parts))    
            s.SendData(tx_messages, self.comm_port)
            return FunctionResult.SUCCESS
        except Exception as e:
            print(f"An error occurred: {e}")
            return FunctionResult.FAILURE

    def reset(self):
        """ 
        reset hand to factory defaults.

        Returns:
            FunctionResult in dexhand.py        
        """ 
        tx_messages = struct.pack('>B', 0x26)
        return self.Send_Messages(tx_messages, port=self.comm_port)


    def calibration(self):
        """
        HandCalibration, parameters reset to default
        Returns:
            FunctionResult in dexhand.py        
        """
        tx_messages = struct.pack('>BB', 0x01, 0x01)
        return self.Send_Messages(tx_messages, port=self.ctrl_port)



    def reset_pid_default(self):
        """ 
        set pid to default values.
        Returns:
            FunctionResult in dexhand.py        
        """ 
        tx_messages = struct.pack('>BB', 0x01, 0x0b)
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_angle(self, id, angle):
        """
        SetControlConfig (Position mode) : set Position move for one finger /all fingers 
        
        Limit:[0 - 12]
        Args:
            one finger :
            id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            angle Params : angle

            multi finger :
            id : 0 (default)
            angle Params : List [index angle, middle angle, ring angle, little angle, thumb-1 angle, thumb-2 angle]

        Returns:
            FunctionResult in dexhand.py  
        """
        if id != 0:
            tx_messages = struct.pack('>BBBBf', 0x01, 0x02, 0x00, int(id), float(angle))
        else:
            tx_messages = struct.pack('>BBBBffffff', 0x01, 0x11, 0x00, 0x00, float(angle[0]), float(angle[1]), float(angle[2]), float(angle[3]), float(angle[4]), float(angle[5]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_velocity(self, id, velocity):
        """
        SetControlConfig (velocity mode) : set velocity move for one finger /all fingers 
        Limit:[0 - 10000]
        
        Args:
            one finger :
                one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
                velocity Params : velocity     

            multi finger :
                id : 0 (default)
                velocity Params : List [index velocity, middle velocity, ring velocity, little velocity, thumb-1 velocity, thumb-2 velocity]

        Returns:
            FunctionResult in dexhand.py  
        """
        if id != 0:
            tx_messages = struct.pack('>BBBBf', 0x01, 0x03, 0x00, int(id), float(velocity))
        else:
            tx_messages = struct.pack('>BBBBffffff', 0x01, 0x12, 0x00, 0x00, float(velocity[0]), float(velocity[1]), float(velocity[2]), float(velocity[3]), float(velocity[4]), float(velocity[5]))

        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_current(self, id, current):
        """
        SetControlConfig (current mode) : set current move for one finger /all fingers 
        Limit:[0 - 1000]
        
        Args:
            one finger :
                id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
                current Params : current       
            
            multi finger :
                id : 0 (default)
                current Params : List [index current, middle current, ring current, little current, thumb-1 current, thumb-2 current]
                             
        Returns:
            FunctionResult in dexhand.py  
        """
        if id != 0:
            tx_messages = struct.pack('>BBBBf', 0x01, 0x04, 0x00, int(id), float(current))
        else:
            tx_messages = struct.pack('>BBBBffffff', 0x01, 0x13, 0x00, 0x00, float(current[0]), float(current[1]), float(current[2]), float(current[3]), float(current[4]), float(current[5]))

        return self.Send_Messages(tx_messages, port=self.ctrl_port)



    def set_pos_pid(self, id, pid):
        """
        SetPIDParams (Position mode) : set PID Position control config params 
        Limit:[0 - 12]
        
        Args:
            one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            PID Params : [p, i, d]

        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBfff', 0x01, 0x05, 0x00, id, float(pid[0]), float(pid[1]), float(pid[2]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_velocity_pid(self, id, pid):
        """
        SetPIDParams (velocity mode) : set PID velocity control config params 
        Limit:[0 - 10000]

        Args:
            one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            PID Params : [p, i, d]
            
        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBfff', 0x01, 0x06, 0x00, id, float(pid[0]), float(pid[1]), float(pid[2]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_current_pid(self, id, pid):
        """
        SetPIDParams (current mode) : set PID current control config params 
        Limit:[0 - 1000]

        Args:
            one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            PID Params : [p, i, d]
            
        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBfff', 0x01, 0x07, 0x00, id, float(pid[0]), float(pid[1]), float(pid[2]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_pos_limit(self, id, limit):
        """
        SetLimitedParams (Position mode) : set position control config params 
        Limit:[0 - 12]

        Args:
            one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            limit : [angle_low, angle_high]

        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBff', 0x01, 0x08, 0x00, id, float(limit[0]),  float(limit[1]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)



    def set_velocity_limit(self, id, limit):
        """
        SetLimitedParams (velocity mode) : set velocity control config params 
        Limit:[0 - 10000]

        Args:
            one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            limit : max velocity

        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBf', 0x01, 0x09, 0x00, id, float(limit))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_current_limit(self, id, limit):
        """
        SetLimitedParams (Current mode) : set Current control config params 
        Limit:[0 - 1000]

        Args:
            one finger id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            limit : max current

        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBf', 0x01, 0x0a, 0x00, id, float(limit))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_pwm(self, target):
        """
        SetControlConfig (PWM mode) : set PWM move for one finger 
        Limit : [ -200 - 200 ]
        Args:
            List []: 6 fingers PWM Params 
            [index_pwm, middle_pwm, ring_pwm, little_pwm, thumb-1_pwm, thumb-2_pwm]
        
        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBffffff', 0x01, 0x14, 0x00, 0x00, float(target[0]), float(target[1]), float(target[2]), 
                                float(target[3]), float(target[4]), float(target[5]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)


    def set_force(self, force):
        """
        Set force through voltage
        Args:
            List []: 6 fingers force Params 
            [index_force, middle_force, ring_force, little_force, thumb-1_force, thumb-2_force]
        
        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBffffff', 0x01, 0x15, 0x00, 0x00, float(force[0]), float(force[1]), float(force[2]), float(force[3]), float(force[4]), float(force[5]))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)
    


    def set_pd_control(self, id, target, w):
        """
        SetControlConfig (PD control mode)
        target limit : [0 - 12]
        Args:
            id : [1: index, 2: middle, 3: ring, 4: little, 5: thumb-1, 6: thumb-2]
            target : position 
            w : omega         
        Returns:
            FunctionResult in dexhand.py  
        """
        tx_messages = struct.pack('>BBBBff', 0x01, 0x21, 0x00, int(id), float(target), int(w))
        return self.Send_Messages(tx_messages, port=self.ctrl_port)



    def clean_error(self):
        """
        clean errors of current or  Electric Cylinder.

        Returns:
            FunctionResult in dexhand.py        
        """ 
        tx_messages = struct.pack('>BBBB', 0x01, 0xf1, 0x00, 0x00)
        return self.Send_Messages(tx_messages, port=self.ctrl_port)
